from typing import TYPE_CHECKING, Annotated

from fastapi import Depends

from typing import Callable


class Stub:  # noqa: F811
    def __init__(self, dependency: Callable, **kwargs):
        self._dependency = dependency
        self._kwargs = kwargs

    def __call__(self):
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        if isinstance(other, Stub):
            return (
                self._dependency == other._dependency and self._kwargs == other._kwargs
            )
        else:
            if not self._kwargs:
                return self._dependency == other
            return False

    def __hash__(self):
        if not self._kwargs:
            return hash(self._dependency)
        serial = (
            self._dependency,
            *self._kwargs.items(),
        )
        return hash(serial)


if TYPE_CHECKING:
    from typing import Union as FastDIStub

else:

    class FastDIStub:
        def __class_getitem__(cls, item):
            return Annotated[item, Depends(Stub(item))]



from typing import TYPE_CHECKING, Annotated

from fastapi import Depends


if TYPE_CHECKING:
    from typing import Union as FastDI

else:

    class FastDI:
        def __class_getitem__(cls, item):
            return Annotated[item, Depends()]



from typing import Callable

from fast_di.collector.base import BaseDependencyCollector, DependencyT, FactoryT


class DependencyCollector(BaseDependencyCollector):
    def __init__(self) -> None:
        super().__init__()

    def factory(self, dependency: DependencyT) -> Callable[[FactoryT], None]:
        def decorator(factory: FactoryT) -> None:
            self.add_factory(dependency, factory)

        return decorator

    def singleton(
        self, dependency: DependencyT, is_async: bool = False
    ) -> Callable[[FactoryT], None]:
        def decorator(factory: FactoryT) -> None:
            self.add_singleton(dependency, factory, is_async)

        return decorator



from fastapi import FastAPI

from fast_di import DependencyCollector


def init_dependencies(app: FastAPI, *collectors: "DependencyCollector") -> None:
    if not collectors:
        raise RuntimeError("No collectors provided")

    for collector in collectors:
        for dependency, factory in collector.dependencies.items():
            app.dependency_overrides.update({dependency: factory})



from functools import wraps
from typing import Callable, TypeVar

DependencyT = TypeVar("DependencyT")
FactoryT = TypeVar("FactoryT", bound=Callable)


class BaseDependencyCollector:
    def __init__(self) -> None:
        self.dependencies: dict[DependencyT, FactoryT] = {}

    def add_factory(self, dependency: DependencyT, factory: FactoryT) -> None:
        self.dependencies[dependency] = factory

    def add_singleton(
        self, dependency: DependencyT, factory: FactoryT, is_async: bool = False
    ) -> None:
        self.dependencies[dependency] = self.__create_singleton(factory, is_async)

    def __create_singleton(
        self, factory: FactoryT, is_async: bool
    ) -> Callable[..., FactoryT]:
        singleton_instance = None

        @wraps(factory)
        async def async_wrapper(*args, **kwargs) -> FactoryT:
            nonlocal singleton_instance

            if singleton_instance is None:
                singleton_instance = await factory(*args, **kwargs)

            return singleton_instance

        @wraps(factory)
        def sync_wrapper(*args, **kwargs) -> FactoryT:
            nonlocal singleton_instance

            if singleton_instance is None:
                singleton_instance = factory(*args, **kwargs)

            return singleton_instance

        return async_wrapper if is_async else sync_wrapper



from fast_di.depends.depends import FastDI
from fast_di.depends.depends_stub import FastDIStub
from fast_di.init_dependencies import init_dependencies
from fast_di.collector.collector import DependencyCollector

__all__ = ["Stub", "FastDI", "FastDIStub", "DependencyCollector", "init_dependencies"]

__version__ = "0.0.2"
__author__ = "Abrekov Alim"
